# -*- coding: utf-8 -*-

# imprint: a program for creating documents from data and content templates
#
# Copyright (C) 2019  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding


"""
Contains most of the code that actually runs imprint.

The this module is executed publicly by the `imprint` script.
"""

# Set up matplotlib environment before doing anything else.
import matplotlib
matplotlib.use('ps')  # This line currently has some OS-specific implications

import logging
from os.path import abspath, join, split, splitext
from pathlib import Path
from pprint import pformat
from warnings import warn
from xml import sax

from docx import Document
from docx.opc.exceptions import PackageNotFoundError

from haggis.files import ensure_extension
from haggis.load import module_as_dict, load_object
from haggis.logs import configure_logger
from haggis.os import chdir_context

from .core.parsers import ReferenceProcessor, TemplateProcessor
from .core.tags import tag_registry
from .core.utilities import substitute_headers_and_footers, trigger_fail_state


#: Private logger for this module
_logger = logging.getLogger(__name__)


def load_extensions(config):
    """
    Loads all the user-defined
    :py:class:`~imprint.core.tags.TagDescriptor`\ s
    laid out in the config.

    The key ``tags`` is expected to contain a mapping of custom tag
    names to either tag descriptor objects, or the fully qualified names
    of such objects, which will be imported in the usual manner.

    Here is an example of some possibilities for ``tags``::

        tags = {
            'custom-tag1': 'repgen.demo.custom1',
            'custom-tag2': custom2,
        }

    The non-string values in tags will be registered in
    :py:data:`imprint.core.tags.tag_registry` directly. String values are
    treated as a package/module path followed by an object name, all dot
    delimited. The object will be imported and registered. The
    package/module that contains it must be on the path.
    """
    for tag, descriptor in config.get('tags', {}).items():
        if isinstance(descriptor, str):
            descriptor = load_object(descriptor)
        tag_registry[tag] = descriptor


def create_output_document(config, fail='warn'):
    """
    Create an output document object based on the configuration.

    The input :ref:`configuration-docx` will contain any styles that the
    output document will use. If the stub has any content, that content
    will appear before the generated content of the to the output
    document.

    The name of the input file is expected to be found in the `config`
    dictionary under :ref:`keywords-system-input_docx`. If the key is
    missing or the value is `None`, a new document is created. If the
    requested stub document is not found, one of the following occurs
    depending on the value of `fail`:

      - ``'ignore'``: Silently create and return an empty document. All
        the styles named in the :ref:`configuration-xml` must match the
        ones contained in the :ref:`configuration-docx`. 
      - ``'warn'``: Raise a warning and return an empty document.
      - ``'raise'``: Raise an :py:exc:`IOError`.

    Any other value for `fail` will trigger a :py:exc:`ValueError`.
    """
    doc_name = config.get('input_docx', None)
    try:
        return Document(doc_name)
    except PackageNotFoundError:
        trigger_fail_state(fail, f'Document file {doc_name} not found',
                           errorClass=IOError)
        return Document()


def save_output_document(document, config):
    """
    Save the specified document to a location given in the
    configuration.

    Two keys are used: :ref:`keywords-system-output_docx` and
    :ref:`keywords-system-overwrite_output`. The former is the file
    name. The latter determines what happens if the file already exists:

      - ``'raise'``: Raise an error.
      - ``'rename'``: Display a prompt and request that the user type in a
        new file name.
      - ``'silent'``: Ignore the issue and overwrite.
      - ``'warn'``: Ignore the issue, overwrite, but raise a warning
        nonetheless.

    If :ref:`keywords-system-overwrite_output` is missing, ``'raise'``
    is the default behavior.
    """
    output_docx = config['output_docx']
    overwrite_output = config.get('overwrite_output', 'raise').casefold()

    if overwrite_output not in ('raise', 'rename', 'warn', 'silent'):
        raise ValueError(
            'Invalid value for config key "overwrite_output". '
            'Must be one of ("raise", "rename", "silent", "warn")'
        )

    mode = 'wb' if overwrite_output == 'silent' else 'xb'

    try:
        output_file = open(output_docx, mode)
    except FileExistsError as err:
        if overwrite_output == 'raise':
            raise
        elif overwrite_output == 'rename':
            while True:
                path, file = split(output_docx)
                base_name, ext = splitext(file)
                def_name = base_name + '-1' + ext
                output_file = input('Refusing to overwrite file.\n'
                                    f'Enter a new name [{def_name}]: ')
                if not output_file.strip():
                    output_file = def_name
                output_file = join(path, output_file)
                try:
                    output_docx = open(output_file, mode)
                except FileExistsError:
                    pass
                else:
                    break
        elif overwrite_output == 'warn':
            warn(f'Overwriting existing file {output_docx!r}', RuntimeWarning)
            output_file = output_docx

    document.save(output_file)
    return output_docx


def post_process_document(file_name, config):
    """
    Apply final touchup to the output document after it has been saved.

    The document object can be assumed to have been destroyed at this
    point. The only access to it is through `file_name`. The current
    implementation filters the headers and footers of the document and
    does a keyword repacement on them.
    """
    substitute_headers_and_footers(file_name, config)


def build_document(config):
    """
    Output a document based on the specifed configuration.

    This is the main method of the :py:mod:`imprint` package. It creates
    the output by using instructions from the :ref:`configuration-xml`
    to append formatted content to the :ref:`configuration-docx`. The
    configuration dictionary directs both the I/O as well as keyword
    parameter replacements in the XML.
    """
    document = create_output_document(config)
    xml_template = abspath(config['input_xml'])

    ref_proc = ReferenceProcessor(config.get('caption_counter_depth', 1))
    sax.parse(xml_template, ref_proc)
    references = ref_proc.state.references
    _logger.debug('Anchor Map:\n%s', references)
    sax.parse(xml_template, TemplateProcessor(config, document, references))
    file_name = save_output_document(document, config)
    post_process_document(file_name, config)


def run(config_name):
    """
    Run the driver for a named configuration file.

    This wrapper loads the configuration and sets up logging before
    calling :py:func:`build_document` in earnest. All of the
    :ref:`logging-configuration` keywords are used here except
    :ref:`keywords-system-log_images`.
    """
    def get_lvl(name, default):
        level = config.get(name, default)
        if isinstance(level, str):
            level = level.upper()
        return level

    config_path = Path(config_name)
    with chdir_context(config_path.parent):
        config = module_as_dict(
            config_path.name, name='config', include_var='includes'
        )
        log_file = config.get('log_file', None)
        if log_file is True:
            log_file = ensure_extension(config['output_docx'], '.log',
                                        partial_policy='replace')
        configure_logger(
            log_file, get_lvl('file_level', 'NOTSET'),
            config.get('log_stderr', False), get_lvl('stderr_level', 'ERROR'),
            config.get('log_stdout', True), get_lvl('stdout_level', 'WARNING'),
            config.get('log_format', None)
        )
        _logger.debug('Loaded config:\n%s',
                      pformat(config, indent=4, width=255))
        load_extensions(config)
        build_document(config)



# -*- coding: utf-8 -*-

# imprint: a program for creating documents from data and content templates
#
# Copyright (C) 2019  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding


"""
:py:mod:`imprint.core.utilities` containins general utilities to help
the engine create and process docx files.

The configuration loaders in this module are potentially suitable for
inclusion in the `haggis`_ library.


.. include:: /link-defs.rst
"""

__all__ = [
    'aggressive_strip', 'check_fail_state', 'trigger_fail_state',
    'get_handler', 'load_callable', 'substitute_headers_and_footers',
]


from datetime import datetime
from inspect import isclass, ismodule
import logging
import re
from warnings import warn
from zipfile import ZipFile

from haggis import logs, Sentinel
from haggis.files.zip import filter as filter_zip

from . import encoding


#: Private logger for this module
_logger = logging.getLogger(__name__)


#: A list of the valid options for :py:func:`check_fail_state` and
#: :py:func:`trigger_fail_state`.
_fail_options = ('raise', 'warn', 'ignore')


logs.add_logging_level('XTRACE', 2, exc_info=True)


def check_fail_state(fail):
    """
    Verify that `fail` is one of the valid options ``{'raise', 'warn',
    'ignore'}``.

    Raise a :py:exc:`ValueError` if it is not.
    """
    if fail not in _fail_options:
        raise ValueError(f'fail mode must be in {_fail_options}')


def trigger_fail_state(fail, msg,
                       error_class=ValueError, warn_class=UserWarning):
    """
    React to a failure according to the value of ``fail``:

      - ``'ignore'``: Do nothing
      - ``'warn'``: Raise a warning with message `msg` and class
        `warn_class` (:py:exc:`UserWarning` by default).
      - ``'raise'``: Raise an error with message `msg` and class
        `error_class` (:py:exc:`ValueError` by default).

    Any other value of `fail` triggers a :py:exc:`ValueError`.
    """
    check_fail_state(fail)
    if fail == 'raise':
        raise error_class(msg)
    elif fail == 'warn':
        warn(msg, warn_class, stacklevel=2)
    else:  # Make ignore explicit
        pass


def get_handler(handler_name):
    """
    Load the named :ref:`plugin handler <plugins-handlers>`.

    Handlers are callables that take an object ID and configuration
    dictionary and generate content for a specific tag like
    :ref:`xml-spec-tags-figure`, :ref:`xml-spec-tags-table` or
    :ref:`xml-spec-tags-string`.

    If the handler is not found as-is, the :py:mod:`imprint.handlers`
    package is prefixed to `handler_name` since that is where all
    built-in handlers live.
    """
    return load_callable(handler_name, 'imprint.handlers')


def load_callable(name, package_prefix=None, magic_module_attribute=Sentinel,
                  instantiate_class=False):
    """
    Retrieve an arbitrary callable from a module

    The input may be one of six things:

     1. A module with a `magic_module_attribute` that contains the
        callable.
     2. A callable that implements the correct interface.
     3. The name of a module containing the `magic_module_attribute`.
     4. The name of a callable.
     5. The name of a module in the `package_prefix` package.
     6. The name of a callable in the `package_prefix` package.

    The correct thing is identified as leniently as possible and
    returned. The returned object is not guaranteed to be the correct
    thing, just to pass very cursory inspection (e.g., modules must have
    the magic attribute and any other objects must be callable)

    Items 1, 3, 5 are not possible if `magic_module_attribute` is not
    specified. Items 5, 6 are not possible if `package_prefix` is not
    specified.

    This method has one special case. If the object found is a class with a
    no-arg `__init__` method and a `__call__` method, an instance rather than
    the class object is returned. Note that class objects themselves are
    callable, so if you specify a class without a no-arg `__init__` method or
    without a `__call__` method, make sure that `__init__` has the signature
    you require and returns the object that you expect.
    """
    def check_callable(candidate):
        """
        Check if an object is callable and return it if it is.

        If `candidate` is not callable, return :py:obj:`None`. If
        `instantiate_class` is set and `candidate` is a class object
        with a no-arg `__init__` method that initializes callable
        instances, an instance will be returned instead of the class.
        """
        if instantiate_class and isclass(candidate):
            try:
                instance = candidate()
            except TypeError:
                _logger.debug(
                    'Importing "%s.%s" as class: does not have a no-arg '
                    'constructor', candidate.__module__, candidate.__qualname__
                )
            else:
                if callable(instance):
                    _logger.debug('Importing callable instance of "%s.%s"',
                                  candidate.__module__, candidate.__qualname__)
                    candidate = instance
                else:
                    _logger.debug(
                        'Importing "%s.%s" as class: instance obtained, but '
                        'is not callable', candidate.__module__,
                        candidate.__qualname__
                    )
        elif callable(candidate):
            _logger.debug(
                'Importing "%s" as regular callable', candidate.__name__
                    if hasattr(candidate, '__name__') else repr(candidate)
            )
        else:
            candidate = None
        return candidate

    def load_candidate(modname, callname=None):
        if not callname and callname is not magic_module_attribute:
            modpath = modname.split('.')
            if len(modpath) == 1:
                callname = ''
                modpath = modname
            else:
                callname = modpath[-1]
                modpath = '.'.join(modpath[:-1])

        if callname:
            try:
                mod = __import__(modpath, None, None, callname)
            except ModuleNotFoundError as e:
                _logger.trace(
                    'Unable to import "%s" no such module: %s', modname, e
                )
            except ImportError as e:
                _logger.trace(
                    'Unable to import "%s" due to non-syntax error in module: '
                    '%s', modname, e
                )
            except SyntaxError as e:
                _logger.trace(
                    'Unable to import "%s" due to syntax error: %s', modname, e
                )
            else:
                if hasattr(mod, callname):
                    return check_candidate(mod, callname)
                else:
                    _logger.trace('Unable to import "%s": module has no '
                                  'such attribute', modname)

        return None

    load_module = lambda modname: load_candidate(modname,
                                                 magic_module_attribute)

    def check_candidate(mod, attr):
        candidate = getattr(mod, attr)  # Raises AttributeError
        if check_callable(candidate) is None:
            raise TypeError(f'{mod.__name__}.{attr} found but is not callable')
        return candidate

    # 1: Actual module containing callable `magic_module_attribute`
    if ismodule(name):
        if magic_module_attribute:
            return check_candidate(name, magic_module_attribute)
        else:
            raise ValueError('Magic attribute name not specified for module')

    # 2: Callable object
    if check_callable(name) is not None:
        return name

    # All following operations done on strings
    if not isinstance(name, str):
        raise TypeError('requested object must be module, callable or string')

    prefixed_name = package_prefix + '.' + name
    for item in (name, prefixed_name):
        for loader in (load_module, load_candidate):
            candidate = loader(item)
            if candidate is not None:
                return candidate

    raise TypeError(f'{name} ({type(name)}) does not correspond '
                    'to a known object')


def substitute_headers_and_footers(doc_file_name, keywords):
    """
    Perform a keyword replacement on all valid newstyle format strings
    in the header and footer XML of a word document.

    This operation is currently done by treating the XML as if it was a
    giant string. The assumption is valid but hacky, since format-like
    strings delimited by '{}' are unlikely to appear anywhere outside
    ``<w:t>`` tags.
    """
    with ZipFile(doc_file_name, 'r') as z:
        file_list = [name for name in z.namelist()
                            if name.startswith('word/footer') or
                               name.startswith('word/header')]

    _logger.debug('Replacing keywords in the following '
                  'headers and footers: %s', file_list)

    if 'date' in keywords:
        formatter = str.format
    else:
        formatter = lambda fmt, **kwds: str.format(fmt, date=datetime.now(),
                                                   **kwds)

    def kwd_replace(string):
        """
        Accept a string of bytes or characters and replace every
        apparent occurrence of a keyword.

        Failed replacements are left alone in the string.

        .. todo::

           The correct way to implement this is to parse the XML and
           only accept replacements between ``<w:t>`` tags. The issue is
           of course that there are too many breaks between ``<w:t>``
           tags.
        """
        if not isinstance(string, str):
            string = string.decode(encoding)

        def try_replace(matcher):
            format_string = matcher.group()
            try:
                replacement = formatter(format_string, **keywords)
            except Exception as e:
                replacement = format_string
                _logger.trace('Failed to replace "%s": Most likely caused by '
                              'no matches in keywords.', format_string)
            else:
                _logger.trace('Replacing "%s" with "%s"',
                              format_string, replacement)

            return replacement

        return re.sub('{[^}]*}', try_replace, string)

    filter_zip(doc_file_name, *file_list, filter=kwd_replace)


def aggressive_strip(string):
    """
    Split a string along newlines, strip surrounding whitespace on
    each line, and recombine with a single space in place of the
    newlines.
    """
    return ' '.join(filter(None, map(str.strip, string.splitlines())))



# -*- coding: utf-8 -*-

# imprint: a program for creating documents from data and content templates
#
# Copyright (C) 2019  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding


"""
:py:mod:`imprint.handlers.table.tables` contains basic built-in
:ref:`plugins-handlers` for inserting simple tables into a document.

All the handlers in this module are compatible with the
:ref:`plugin interface <plugins-tables-signature>` used by the
:ref:`xml-spec-tags-table` tag.

This module contains handlers that are only enabled when some optional
dependencies are installed. See :py:mod:`imprint.handlers.table` for
more information.
"""

import logging
from shutil import copyfile

try:
    import pandas as pd
except ImportError:
    from sys import stderr
    print('It appears that you do not have pandas installed. '
          'Try one of the following:\n\n'
          '    pip install imprint[all]\n\n'
          'OR\n\n'
          '    pip install pandas', file=stderr)
    raise

from haggis.files.docx import merge_row
from haggis.recipes import dict_select

from . import fixed_width_table
from ..utilities import get_file


__all__ = ['CSVFile', 'DataFrame']


#: Private logger for this module
_logger = logging.getLogger(__name__)


def CSVFile(config, kwds, doc, style, *, image_log_name=None):
    """
    Generate a table from a CSV file.

    Files can be parsed with or without header information. By default,
    the first row and column are both considered to be headers.

    If image logging is enabled (``image_log_name`` is not falsy), the
    contents of the file is copied to the specified output CSV file. A
    ``.csv`` extension will be appended to the name.

    The following :ref:`plugins-data-configuration` keys are used:

        file
            The (mandatory) file name.
        formatted
            Whether or not ``file`` is a format string that has keyword
            replacements in it. Defaults to truthy. Set to falsy if
            the name contains random opening braces.
        row_header
            Whether or not to treat the first row of the CSV as a table
            header. Optional, defaults to :py:obj:`True`.
        col_header
            Whether or not to treat the first column of the CSV as a
            table header. Optional, defaults to :py:obj:`True`.
        heading
            A text label that will appear across the entire first row of
            the table if specified.
        alignment
            The direction in which the table is to be aligned on the
            page. One of { ``'<'``, ``'^'``, ``'>'`` }. Defaults to
            center (``'^'``).
        width
            The size of the table, either as a number in inches, or a
            string representing a number with optional units. See
            :py:func:`haggis.files.docx.str2length` for a full list of
            units. Defaults to 7.0 inches.
        col_widths
            A sequence of column widths. If the sequence is longer than
            the actual number of columns, extra elements will be
            silently ignored. If shorter, the remaining *width* will be
            divided equally between remaining columns. If the sum of the
            used portion of *col_widths* is greater than *width*, an
            error will be raised. The elements of the sequence can be
            numbers in inches, or strings with optional units, just like
            *width*.
        row_height : None or length-spec or sequence[length-spec]
            The height to assign to each row. :py:obj:`None` is the
            default, meaning to let the rows adjust themselves. If a
            single number or numerical string with optional units, it
            will be applied to all the newly created rows. If a
            sequence, only the minimum of ``rows`` and the length of the
            sequence will be affected. A sequence may contain any of the
            values allowed as scalars (including :py:obj:`None`).

    Notes
    -----
    - Whether or not the headings appear as actual headings depends on
      the requested table style.
    - This plugin uses the `pandas`_ library rather than the builtin
      :py:mod:`csv` module to ensure data shape and type normalization.
    """
    input_path = get_file(config, kwds)
    _logger.trace('Loading datarame from %r', input_path)

    row_header = config.get('row_header', True)
    col_header = config.get('col_header', True)

    header = 0 if row_header else None
    index_col = 0 if col_header else None

    data_frame = pd.read_csv(input_path, header=header, index_col=index_col)

    df_config = {
        'data': data_frame,
        'header': row_header,
        'index': col_header,
    }
    df_config.update(dict_select(config, ('heading', 'alignment', 'width',
                                          'col_widths', 'row_height')))

    if image_log_name:
        copyfile(input_path, image_log_name + '.csv')

    return DataFrame(df_config, kwds, doc, style, image_log_name=None)


def DataFrame(config, kwds, doc, style, *, image_log_name=None):
    """
    Generate a table from a `pandas`_ :py:class:`~pandas.DataFrame`.

    Files can be parsed with or without header information. By default
    both the column names and the index are considered headers. Headers
    can be omitted on either side.

    If image logging is enabled (``image_log_name`` is not falsy), the
    contents of the dataframe is dumped to the specified file in CSV
    format. A ``.csv`` extension will be appended to the name.

    The following :ref:`plugins-data-configuration` keys are used:

        data
            The (mandatory) :py:class:`~pandas.DataFrame` to display.
        header
            Whether or not to use the column names as a table header.
            Defaults to :py:obj:`True`.
        index
            Whether or not to use the index as a table header. Defaults
            to :py:obj:`True`. If both *index* and *header* are truthy,
            The name of the index will appear in the upper-leftmost cell
            of the generated table.
        heading
            A text label that will appear across the entire first row of
            the table if specified.
        alignment
            The direction in which the table is to be aligned on the
            page. One of { ``'<'``, ``'^'``, ``'>'`` }. Defaults to
            center (``'^'``).
        width
            The size of the table, either as a number in inches, or a
            string representing a number with optional units. See
            :py:func:`haggis.files.docx.str2length` for a full list of
            units. Defaults to 7.0 inches.
        col_widths
            A sequence of column widths. If the sequence is longer than
            the actual number of columns, extra elements will be
            silently ignored. If shorter, the remaining *width* will be
            divided equally between remaining columns. If the sum of the
            used portion of *col_widths* is greater than *width*, an
            error will be raised. The elements of the sequence can be
            numbers in inches, or strings with optional units, just like
            *width*.
        row_height : None or length-spec or sequence[length-spec]
            The height to assign to each row. :py:obj:`None` is the
            default, meaning to let the rows adjust themselves. If a
            single number or numerical string with optional units, it
            will be applied to all the newly created rows. If a
            sequence, only the minimum of ``rows`` and the length of the
            sequence will be affected. A sequence may contain any of the
            values allowed as scalars (including :py:obj:`None`).

    Notes
    -----
    Whether or not the headings appear as actual headings depends on the
    requested table style.
    """
    data = config['data']

    label = config.get('heading')
    heading = bool(label)

    header = bool(config.get('header', True))
    index = bool(config.get('index', True))

    offset = header + heading

    rows = data.shape[0] + offset
    cols = data.shape[1] + index

    kwargs = dict_select(config, ('alignment', 'width',
                                  'col_widths', 'row_height'))
    table = fixed_width_table(doc, style, rows, cols, **kwargs)

    if heading:
        table.cell(0, 0).text = label
        merge_row(table, row=0)
    if header:
        for col, name in enumerate(data.columns, start=index):
            table.cell(heading, col).text = format(name)
    if index:
        for row, val in enumerate(data.index, start=offset):
            table.cell(row, 0).text = format(val)
        if header and data.index.name:
            table.cell(heading, 0).text = format(data.index.name)

    for r_t, r_df in enumerate(range(data.shape[0]), start=offset):
        for c_t, c_df in enumerate(range(data.shape[1]), start=index):
            table.cell(r_t, c_t).text = format(data.iloc[r_df, c_df])

    if image_log_name:
        data.to_csv(image_log_name + '.csv')



# -*- coding: utf-8 -*-

# imprint: a program for creating documents from data and content templates
#
# Copyright (C) 2019  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding


"""
:py:mod:`imprint.handlers.figure.images` contains basic built-in
:ref:`plugins-handlers` for inserting images into a document. All the
handlers in this module are compatible with the
:ref:`plugin interface <plugins-figures-signature>` used by the
:ref:`xml-spec-tags-figure` tag.
"""

import errno
from io import BytesIO
import logging
from os import strerror
from os.path import exists, samefile, splitext
from shutil import copyfile, copyfileobj

from haggis.files.pdf import pdf_to_image

from ..utilities import get_file


__all__ = ['ImageFile',]


#: Private logger for this module
_logger = logging.getLogger(__name__)


def ImageFile(config, kwds, output=None):
    """
    Generate `python-docx`_ compatible images from image files.

    Copy image files as-is, or load them into memory. Output must be
    to a file of the same type as the input (except for PDFs): no
    conversion is done, only direct copy. PDFs (identified by the
    ``'.pdf'`` extension) get special handling to convert them into
    usable images.

    The following :ref:`plugins-data-configuration` keys are used:

        file
            The (mandatory) file name containing the image.
        formatted
            Whether or not ``file`` is a format string that has keyword
            replacements in it. Defaults to truthy. Set to falsy if
            the name contains random opening braces.

    Notes
    -----
    Using this plugin with PDF files requires the `poppler`_ library
    mentioned in the :ref:`dependencies-plugins-external`.

    .. include:: /link-defs.rst
    """
    input_path = get_file(config, kwds)

    if not exists(input_path):
        # http://stackoverflow.com/a/36077407/2988730
        # has instructions on how to raise FNFE properly
        raise FileNotFoundError(errno.ENOENT, strerror(errno.ENOENT),
                                input_path)
    _, input_ext = splitext(input_path)

    # PDFs get special treatment
    if input_ext == '.pdf':
        if isinstance(output, str):
            if not output.endswith('.png'):
                output += '.png'
            desc = '"%s"' % output
        elif output is None:
            desc = 'in-memory PNG'
        else:
            desc = 'existing PNG stream'

        _logger.info('Converting PDF "%s" to %s', input_path, desc)
        output = pdf_to_image(input_path, output, format='png')

    elif isinstance(output, str):
        # file is a string path name
        if not output.endswith(input_ext):
            output += input_ext
        if not exists(output) or not samefile(input_path, output):
            _logger.info('Copying "%s" to "%s"', input_path, output)
            copyfile(input_path, output)
        else:
            _logger.info('Input and output image are the same path: "%s"',
                         input_path)

    else:
        if output is None:
            # Load file into memory
            _logger.trace('Loading "%s" into memory', input_path)
            output = BytesIO()
            rewind = True
        else:
            # file is a file-like object
            _logger.info('Copying "%s" to file-like object', input_path)
            rewind = False

        with open(input_path, 'rb') as f:
            copyfileobj(f, output)

        if rewind:
            output.seek(0)

    return output
